# Web Services controllers
"""Handles the Controller part of MVC for web services. This provides the
WebServicesController class which is the main entry point for people wanting
to use web services."""

import logging
import types
import re
import sys
import traceback

import cherrypy
import dispatch
import cElementTree as et
from genshi.builder import tag, Element

from turbogears import controllers, expose as tgexpose
from turbogears import config
from turbogears.decorator import weak_signature_decorator
from turbogears import validators

from tgwebservices import soap, iconv
from tgwebservices.runtime import register, primitives, \
                                  FunctionInfo, register_complex_type, \
                                  ctvalues

config.update({"tg.empty_flash" : False})

log = logging.getLogger("tgwebservices.controller")

def dict_wrapped_result(func):
    """Wraps a call to a function so that the return value is put into a
    single item dictionary, with the key 'result'.
    
    @param func Function to wrap
    @return a new function
    """
    def d_w_r(*args, **kw):
        return dict(result=func(*args, **kw))
    return d_w_r

_wrong_parameters = re.compile(r".*\(\) takes .* \(\d+ given\)")
_unexpected_parameters = re.compile(r".*\(\) got an unexpected keyword "
                                    "argument .*")

def wsexpose(return_type=basestring):
    """Exposes a method for access via web services. Only one value
    can be returned because many languages only support a single
    return value from methods.
    
    @param return_type the class that will be returned. This is used to
                       help statically typed clients."""
    def entangle(func):
        fi = register(func)
        fi.return_type = return_type
        
        # we'll use the TurboGears version of expose to handle JSON
        # output.
        tgfunc = tgexpose("wsautoxml", accept_format="text/xml",
                          allow_json=True)(dict_wrapped_result(func))
        
        def newfunc(self, **kw):
            if "tg_format" in kw:
                if kw["tg_format"] == "json":
                    cherrypy.request.headers["Accept"] = "text/javascript"
                else:
                    cherrypy.request.headers["Accept"] = "text/xml"
                del kw["tg_format"]
            
            # TurboJson doesn't support application/json at present, but we want
            # to accept that.
            if cherrypy.request.headers.get("Accept", None) == "application/json":
                cherrypy.request.headers["Accept"] = "text/javascript"
            
            request = cherrypy.request
            input_types = fi.input_types
            
            # the _tgws keyword is used when a web service system, such as
            # SOAP, is calling the function rather than a call that is being
            # made via direct URL traversal.
            if "_tgws" in kw:
                webservice = True
                del kw["_tgws"]
            else:
                webservice = False
            
            if "xml_body" in kw:
                kw = iconv.handle_xml_params(kw["xml_body"], input_types)
            elif request.headers.get("Content-Type", "") \
                 .startswith("text/xml"):
                try:
                    clen = int(request.headers.get('Content-Length')) or 0
                    data = request.body.read(clen)
                    body = et.fromstring(data)
                except SyntaxError:
                    raise validators.Invalid("Request XML is invalid", "", 
                                             None)
                kw = iconv.handle_xml_params(body, input_types)
            elif request.headers.get("Content-Type", "") \
                .startswith("application/json"):
                clen = int(request.headers.get('Content-Length')) or 0
                data = request.body.read(clen)
                kw = iconv.handle_json_params(data, input_types)
            else:
                iconv.handle_keyword_params(kw, input_types)
            
            try:
                if webservice:
                    return func(self, **kw)
                else:
                    # when direct traversal is used, call the standard
                    # TurboGears expose mechanism.
                    return tgfunc(self, **kw)
            except TypeError:
                # handle the case where a bogus parameter was sent in to the
                # function
                msg = str(sys.exc_info()[1])
                if (_unexpected_parameters.match(msg) or 
                    _wrong_parameters.match(msg)) and \
                    msg.startswith(func.__name__):
                    raise validators.Invalid(
                        "Unexpected parameter in function call (%s)" % msg, 
                        None, None)
                else:
                    raise
                
        newfunc.exposed = True
        newfunc.__name__ = func.__name__
        newfunc.__doc__ = func.__doc__
        newfunc._ws_func_info = fi
        # a list denotes that the user is returning an array of the type
        # given by the first element in the array
        # we need to keep track of the complex types so that they can be
        # defined as appropriate for the given protocol
        if isinstance(return_type, list):
            rt = return_type[0]
            if rt not in primitives:
                register_complex_type(fi, rt)
        elif return_type not in primitives:
            register_complex_type(fi, return_type)

        return newfunc
    return entangle

def wsvalidate(*args, **kw):
    """Validates and converts incoming parameters. Also registers the 
    parameters used by the method for use by statically typed languages.
    Method parameters can be specified via positional or keyword arguments.
    You should pass in the class used for each parameter."""
    def entangle(func):
        fi = register(func)
        input_types = dict()
        
        # match up the validators with the function parameters
        
        # the validators list doesn't include self, but the parameters list
        # does. So, the parameters list is offset by one higher.
        for i in range(0, len(args)):
            argtype = args[i]
            input_types[fi.params[i]] = argtype
        
        input_types.update(kw)
        fi.input_types = input_types
        return func
    return entangle

class WebServicesController(object):
    """A controller that implements a piece of a web services API."""
    
    def _ws_gather_functions_and_types(self, prefix):
        """Collects all of the functions and types on this controller and the
        controllers beneath it."""
        funcs = dict()
        complex_types = set()
        for key in dir(self):
            if key.startswith("_"):
                continue
            item = getattr(self, key)
            
            # functions are globally registered with Java-style
            # camelCase names
            if prefix:
                newname = prefix+key[0].upper()+key[1:]
            else:
                newname = key
            
            # check for functions and sub-controllers
            if isinstance(item, types.MethodType) and hasattr(item, "_ws_func_info"):
                funcs[newname] = item
                complex_types.update(item._ws_func_info.complex_types)
            elif hasattr(item, "_ws_gather_functions_and_types"):
                morefuncs, moretypes = \
                            item._ws_gather_functions_and_types(newname)
                funcs.update(morefuncs)
                complex_types.update(moretypes)
        return funcs, complex_types
        
    
class WebServicesRoot(WebServicesController):
    """The root of a multi-protocol web service."""
    
    def __init__(self, baseURL, tns=None, typenamespace=None):
        """Constructor for a WebServicesRoot controller.
        
        @param baseURL Sets the URL path for this controller. Some protocols
            need to know this.
        @param tns The SOAP target namespace (defaults to baseURL+soap/)
        @param typenamespace the namespace for the SOAP types (defaults to
            baseURL+soap/types)
        """
        
        if not tns:
            tns = baseURL + "soap/"
        if not typenamespace:
            typenamespace = tns + "types"
        self._ws_baseURL = baseURL
        self._ws_parent = None
        self.soap = soap.SoapController(self, tns, typenamespace)
        self._ws_funcs, self._ws_complex_types = \
            self._ws_gather_functions_and_types("")
        
    def _cp_on_error(self):
        excinfo = sys.exc_info()
        cherrypy.response.body = self._formatexception(excinfo)
    
    @tgexpose("wsautoxml", allow_json=True, accept_format="text/xml")
    def _formatexception(self, excinfo):
        if isinstance(excinfo[1], validators.Invalid):
            return dict(faultcode="Client",
                        faultstring=str(excinfo[1]))
        else:
            return dict(faultcode="Server",
                        faultstring=str(excinfo[1]),
                        debuginfo="\nTraceback:\n%s\n" % 
                            "\n".join(traceback.format_exception(
                                                    *excinfo)))
    



"""Tests for HTTP+XML services"""

import cStringIO as StringIO

import cherrypy
from turbogears import testutil

from tgwebservices.tests.fixtures import *

def test_simple():
    cherrypy.root = MyService("http://foo.bar.baz")
    testutil.create_request("/times2?value=5")
    output = cherrypy.response.body[0]
    print output
    assert output == """<result>10</result>"""

def test_xml_error():
    testutil.create_request("/times2?value=5&foo=1")
    output = cherrypy.response.body[0]
    print output
    assert output == """<result><faultcode>Client</faultcode><faultstring>foo is not a valid parameter (valid values are: ['value'])</faultstring></result>"""

def test_complex_input():
    cherrypy.root = ComplexService("http://foo.bar.baz/")
    request = """<request>
    <person><name>Fred</name><age>22</age></person>
</request>"""
    testutil.create_request("/tenyearsolder", rfile=StringIO.StringIO(request), 
                            method="POST", 
                            headers={"Content-Length" : str(len(request)),
                                     "Content-Type" : 
                                        "text/xml; charset=utf-8"})
    output = cherrypy.response.body[0]
    print output
    assert output == """<result><age>32</age><computed>Hello!</computed>""" \
                     """<name>Fred</name></result>"""


import cStringIO as StringIO

import cherrypy
from turbogears import testutil

from tgwebservices.controllers import WebServicesRoot, wsexpose, wsvalidate
from tgwebservices.tests.fixtures import *

def test_simple():
    cherrypy.root = MyService("http://foo.bar.baz")
    testutil.create_request("/times2?value=5&tg_format=json")
    output = cherrypy.response.body[0]
    print output
    assert output == """{"result": 10}"""

def test_nested_top():
    cherrypy.root = OuterService("http://foo.bar.baz")
    testutil.create_request("/ultimatequestion", 
                            headers=dict(Accept="application/json"))
    output = cherrypy.response.body[0]
    print output
    assert output == """{"result": 42}"""

def test_complex():
    cherrypy.root = ComplexService("http://foo.bar.baz")
    testutil.create_request("/getfancy?tg_format=json")
    output = cherrypy.response.body[0]
    print output
    assert output == """{"result": {"age": 33, "computed": "Hello!", "name": "Mr. Test"}}"""

def test_complex_property():
    cherrypy.root = ComplexService("http://foo.bar.baz")
    testutil.create_request("/getcomprop?tg_format=json")
    output = cherrypy.response.body[0]
    print output
    assert output == """{"result": {"athing": {"age": 55, "computed": "Hello!", "name": "Arnie"}}}"""

def test_error_handling():
    testutil.create_request("/somestrings?foo=1&tg_format=json")
    output = cherrypy.response.body[0]
    print output
    assert output == """{"faultcode": "Client", "faultstring": "foo is not a valid parameter (valid values are: [])"}"""

class BooleanInput(WebServicesRoot):
    @wsexpose(bool)
    @wsvalidate(bool)
    def gotbool(self, truefalse):
        return truefalse

def test_boolean_input():
    def check(test, expected):
        cherrypy.root = BooleanInput("http://foo.bar.baz/")
        testutil.create_request("/gotbool?tg_format=json&truefalse=%s" % test)
        output = cherrypy.response.body[0]
        print output
        assert output == """{"result": %s}""" % expected
    
    tests = [
        ("False", "false"),
        ("True", "true"),
        ("1", "true"),
        ("0", "false"),
        ("false", "false"),
        ("no", "false"),
        ("T", "true"),
        ("F", "false"),
        ("Y", "true"),
        ("N", "false"),
    ]
    for test in tests:
        yield check, test[0], test[1]

def test_complex_input():
    cherrypy.root = ComplexService("http://foo.bar.baz/")
    
    request = """{
        "person" : {"name" : "Fred", "age" : 22}
}"""
    testutil.create_request("/tenyearsolder", rfile=StringIO.StringIO(request), 
                            method="POST", 
                            headers={"Content-Length" : str(len(request)),
                                     "Content-Type" : 
                                        "application/json; charset=utf-8"})
    output = cherrypy.response.body[0]
    print output
    print cherrypy.root.last_person.age
    assert cherrypy.root.last_person.age == 32


"""Runtime registration data."""

import inspect
import types

from turbojson import jsonify

class typedproperty(property):
    def __init__(self, type, *args, **kw):
        self.type = type
        super(typedproperty, self).__init__(*args, **kw)

class unsigned(object):
    """Represents an unsigned integer to the outside world. Within
    Python, it just creates an int."""
    def __new__(cls, value=None):
        if value is not None:
            if value < 0:
                # avoid import loop
                from tgwebservices.base import ClientError
                raise ClientError("Value of %s is not an unsigned integer" %
                                  value)
            return int(value)
        else:
            return int()
        
# These are the primitive types that are generally understood equally well
# by the various protocols.
primitives = set([int, basestring, str, unicode, float, bool, list, dict,
                  long, typedproperty, unsigned])

class FunctionInfo(object):
    """Keeps track of information about a function that is available for
    web services use.
    """
    
    return_type = basestring
    _params = None
    input_types = {}
    
    def __init__(self, params, defaults):
        self.params = params
        
        if defaults:
            # the optional parameters are the ones that are listed in defaults
            # the only optional parameters are at the *end* of the argument
            # list
            optional = [params[i] for i in 
                            range(len(params) - len(defaults), len(params))]
            self.optional = set(optional)
        else:
            self.optional = []
        self.complex_types = set()
    
    def _set_params(self, value):
        # remove the "self" parameter, because it's not relevant for web services
        # declarations
        if value and value[0] == "self":
            value = value[1:]
        self._params = value
    
    params = property(lambda self: self._params, _set_params)
    
    def __repr__(self):
        return "Return type: %s, params: %s" % (self.return_type, self.params)
    
def register(func):
    """Registers a function as available for web services. You can call this
    repeatedly and will get back the same FunctionInfo object each time.
    
    @param func Function to register
    @return a FunctionInfo object with the information saved for this
            function."""
    if not hasattr(func, "_ws_func_info"):
        argspec = inspect.getargspec(func)
        func._ws_func_info = FunctionInfo(argspec[0], argspec[3])
    
    return func._ws_func_info

def ctvalues(cls):
    """Inspects a class for the defined default values. This is used to
    figure out the types of attributes that instances of the class will
    be expected to have (to provide useful information for statically
    typed languages). It will strip out any attributes that start with _
    and any methods.
    
    @param cls class to inspect
    @return list of keys that are likely holders of useful data"""
    return [key for key in dir(cls) if not key.startswith("_") and not 
            isinstance(getattr(cls, key), types.MethodType)]

def _setup_new_type(cls):
    cls._type_dependents = 0
    
    # simplify types that we know about
    @jsonify.jsonify.when("isinstance(obj, cls)")
    def jsonify_cls(obj):
        simplified = dict()
        for key in ctvalues(cls):
            val = getattr(obj, key)
            if val is not None and type(val) not in primitives:
                val = jsonify.jsonify(val)
            simplified[key] = val
        return simplified

def register_complex_type(fi, cls):
    """Remembers a complex type (user-defined class) that has been discovered and 
    also discovers any complex types that are referred  to be the given type.
    Additionally, this will keep track of how many classes depend on a given
    class in the _type_dependencies class variable. This is used as a crude
    mechanism for determining an appropriate order for types to appear in when
    being defined for consumption by clients.
    
    @param fi the FunctionInfo object that will hold the complex types
    @param cls The class to add and inspect"""
    if not cls or cls in fi.complex_types:
        return
    if not hasattr(cls, "_type_dependents"):
        _setup_new_type(cls)
    fi.complex_types.add(cls)
    for key in ctvalues(cls):
        checkitem = getattr(cls, key)
        
        if isinstance(checkitem, list):
            checkitem = checkitem[0]
        if not isinstance(checkitem, type):
            checkitem = checkitem.__class__
        # figure out of this is a complex type. If so, increase
        # its dependents count
        if checkitem != types.NoneType and checkitem not in primitives:
            if not hasattr(checkitem, "_type_dependents"):
                _setup_new_type(checkitem)
            checkitem._type_dependents += 1
            register_complex_type(fi, checkitem)
        


# Release information about TGWebServices

version = "1.1.1"

description = "Multiprotocol Web Services for TurboGears"
long_description = """TurboGears gives you a plain HTTP with JSON return values API for your application for free. This isn't always what you want, though. Sometimes, you don't want to expose all of the data to the web that you need to render your templates. Maybe you need to support a protocol that names the function it's calling as part of what it POSTs such as SOAP or XML-RPC.

TGWebServices provides a super simple API for creating web services that are available via SOAP, HTTP->XML, and HTTP->JSON. The SOAP API generates WSDL automatically for your Python and even generates enough type information for statically typed languages (Java and C#, for example) to generate good client code on their end.

How easy is it?

::

    class Multiplier(WebServicesRoot):

        @wsexpose(int)
        @wsvalidate(int, int)
        def multiply(self, num1, num2):
            return num1 * num2

With this at the root, SOAP clients can find the WSDL file at /soap/api.wsdl and POST SOAP requests to /soap/. HTTP requests to /multiply?num1=5&num2=20 will return an XML document with the result of 100. Add ?tg_format=json (or an HTTP Accept: text/javascript header) and you'll get JSON back.

The great thing about this is that the code above looks like a '''normal Python function''' and doesn't know a thing about web services.

Features
--------

* Easiest way to expose a web services API
* Supports SOAP, HTTP+XML, HTTP+JSON
* Outputs wrapped document/literal SOAP, which is the most widely compatible format
* Provides enough type information for statically typed languages to generate conveniently usable interfaces
* Can output instances of your own classes
* Can also accept instances of your classes as input
* Works with TurboGears 1.0
* MIT license allows for unrestricted use"""
author = "Kevin Dangoor"
email = "dangoor+tgwebservices@gmail.com"
copyright = "Copyright 2006, 2007 Kevin Dangoor, Arbor Networks"

url = "http://tgwebservices.python-hosting.com/"
license = "MIT"



import re
import httplib
import cStringIO as StringIO
import mimetools

from turbogears import testutil, validate
import cherrypy
import cElementTree as et

from tgwebservices.controllers import WebServicesRoot, WebServicesController, \
                                      wsexpose, wsvalidate
from tgwebservices import soap
from tgwebservices.tests.fixtures import *
from tgwebservices import runtime

def run_soap(method, params=""):
    request = """<?xml version="1.0"?>
<soap:Envelope
xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"
soap:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">

  <soap:Body xmlns="http://foo.bar.baz/soap/">
    <%(method)s>
        %(params)s
    </%(method)s>
  </soap:Body>

</soap:Envelope>
""" % dict(method=method, params=params)

    testutil.create_request("/soap/", rfile=StringIO.StringIO(request), 
                            method="POST", 
                            headers={"Content-Length" : str(len(request)),
                                     "Content-Type" : 
                                        "application/soap+xml; charset=utf-8"})
    output = cherrypy.response.body[0]
    return output

def get_wsdl():
    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print output
    return output

def test_basic_call():
    cherrypy.root = MyService("http://foo.bar.baz/")
    
    output = run_soap("times2", "<value>5</value>")
    print output
    root = et.fromstring(output)
    result_elements = list(root.findall(".//{http://foo.bar.baz/soap/types}result"))
    assert len(result_elements) == 1, "Should only have one result"
    elem = result_elements[0]
    print "result:", elem
    assert elem.attrib["{http://www.w3.org/2001/XMLSchema-instance}type"] == \
        "xsd:int", "Expected result to be labeled an int"
    assert elem.text == "10", "Expected result to be 10"

def test_wsdl_with_ints():
    cherrypy.root = MyService("http://foo.bar.baz/")

    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print output
    assert cherrypy.response.headers["Content-Type"] == "text/xml;charset=utf-8"
    root = et.fromstring(output)
    schema_elements = root.findall(".//{http://www.w3.org/2001/XMLSchema}element")
    found = False
    for elem in schema_elements:
        print "FE:", elem
        if "name" in elem.attrib and elem.attrib["name"] == "times2Response":
            found = True
            interior_elem = elem[0][0][0]
            print "response value: %s (%s)" % (interior_elem, 
                                               interior_elem.attrib)
            assert elem[0][0][0].attrib["type"] == "xsd:int"
    assert found, "Should have found a definition for times2Response"

def test_wsdl_with_complex():
    cherrypy.root = ComplexService("http://foo.bar.baz/")
    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print output
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}element")
    found = False
    for elem in schema_elements:
        if "name" in elem.attrib and elem.attrib["name"] == \
                "getfancyResponse":
            found = True
            interior_elem = elem[0][0][0]
            print "response value: %s (%s)" % (interior_elem, 
                                               interior_elem.attrib)
            assert elem[0][0][0].attrib["type"] == "types:FancyValue"
    
    complex_types = root.findall(
                        ".//{http://www.w3.org/2001/XMLSchema}complexType")
    assert found, "Should have found a definition for fancyResponse"
    
    found = False
    found_array = False
    for elem in complex_types:
        if "name" in elem.attrib and elem.attrib["name"] == \
                        "FancyValue":
            found = True
            objattributes = elem[0].getchildren()
            found_name = False
            found_age = False
            found_computed = False
            for attr in objattributes:
                if attr.attrib["name"] == "name":
                    assert attr.attrib["type"] == "xsd:string"
                    found_name = True
                elif attr.attrib["name"] == "age":
                    assert attr.attrib["type"] == "xsd:int"
                    found_age = True
                elif attr.attrib["name"] == "computed":
                    assert attr.attrib["type"] == "xsd:string"
                    found_computed = True
                else:
                    assert False, "Found a %s element which shouldn't "\
                            "be there" % (attr.attrib["name"])
            assert found_name, "Did not find name element"
            assert found_age, "Did not find age element"
            assert found_computed, "Did not find computed element"
        elif "name" in elem.attrib and elem.attrib["name"] == \
            "FoodItem_Array":
            found_array = True
    assert found, "Should have found a delcaration for FancyValue"
    assert found_array, "Should have found the FoodItem_Array"

def test_soap_array():
    cherrypy.root = ComplexService("http://foo.bar.baz/")
    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print output
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}complexType")
    found = False
    stringarray_found = False
    for elem in schema_elements:
        if "name" in elem.attrib and elem.attrib["name"] == \
                "FancyValue_Array":
            found = True
            interior_elem = elem[0]
            assert interior_elem.tag == \
                "{http://www.w3.org/2001/XMLSchema}sequence", \
                "Expected to find a sequence in FancyValue_Array"
            interior_elem = interior_elem[0]
            assert interior_elem.tag == \
                "{http://www.w3.org/2001/XMLSchema}element", \
                "Expected to find element in FancyValue_Array"
            print interior_elem.attrib
            assert interior_elem.attrib["maxOccurs"] == \
                "unbounded", "expected unbounded array"
            assert interior_elem.attrib["nillable"] == \
                "true", "Array should be nillable"
            assert interior_elem.attrib["type"] == \
                "types:FancyValue", \
                "Should have found array type declaration of FancyValue"
        if "name" in elem.attrib and elem.attrib["name"] == \
                "String_Array":
            stringarray_found = True
            interior_elem = elem[0][0]
            assert interior_elem.tag == \
                "{http://www.w3.org/2001/XMLSchema}element", \
                "Expected to find element in String_Array"
            print interior_elem.attrib
            assert interior_elem.attrib["type"] == \
                "xsd:string", \
                "Should have found string type for String_Array"
    
    assert found, "Expected to find the FancyValiue_Array type"
    assert stringarray_found, "Expected to find the String_Array type"
    
    output = run_soap("somestrings")
    root = et.fromstring(output)
    print "SOAP somestrings call output:", output
    result_elements = list(root.findall(".//{http://foo.bar.baz/soap/types}result"))
    assert len(result_elements) == 1
    items = result_elements[0].getchildren()
    assert len(items) == 3, "Expected three items in the array"
    assert items[0].text == "A"
    assert items[1].text == "B"
    assert items[2].text == "C"

def test_nested_complex():
    cherrypy.root = ComplexService("http://foo.bar.baz/")
    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print "WSDL:\n\n"
    print output
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}complexType")
    found = False
    for elem in schema_elements:
        if "name" in elem.attrib and elem.attrib["name"] == "FoodOrder":
            found = True
            elements = elem.getchildren()[0].getchildren()
            assert len(elements) == 2
            foundarray = False
            for e in elements:
                if e.attrib["type"] == "types:FoodItem_Array":
                    foundarray = True
            assert foundarray, "Should have found array of FoodItems"
    assert found, "Should have found FoodOrder type"
    output = run_soap("getorder")
    print "\n\nSOAP:\n\n"
    print output
    assert '<item xsi:type="FoodItem">' in output
    root = et.fromstring(output)
    persontags = root.findall(".//{http://foo.bar.baz/soap/types}person")
    assert len(persontags) > 0
    assert FoodItem._type_dependents == 1

class LocalObject(object):
    name = ""

class LocalController(WebServicesRoot):
    @wsexpose(LocalObject)
    def getobj(self):
        return LocalObject()

def test_complex_types_stay_in_one_controller():
    cherrypy.root = LocalController("http://foo.bar.baz")
    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print "WSDL:\n\n"
    print output
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}complexType")
    for elem in schema_elements:
        assert not ("name" in elem.attrib and elem.attrib["name"] == "FoodOrder"), \
            "Complex type definitions should be limited to controllers that use them"

def test_nested_controllers():
    cherrypy.root = OuterService("http://foo.bar.baz")
    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print "WSDL:\n\n"
    print output
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}complexType")
    found = False
    for elem in schema_elements:
        if "name" in elem.attrib and elem.attrib["name"] == \
                        "FancyValue":
            found = True
    assert found, "Should have found FancyValue from the nested controller"
    operations = set(["ultimatequestion", "innerGetFancy", "innerFooGetFancy",
                      "anotherGetFancy"])
    porttype = root.findall(".//{http://schemas.xmlsoap.org/wsdl/}portType")[0]
    op_elements = porttype.findall(
                          ".//{http://schemas.xmlsoap.org/wsdl/}operation")
    for elem in op_elements:
        name = elem.attrib["name"]
        assert name in operations, "Found unexpected op %s in WSDL" % (name)
        operations.remove(name)
    assert not operations, "Did not find these ops: %s" % (operations)
    
def test_call_to_nested_controller():
    cherrypy.root = OuterService("http://foo.bar.baz")
    output = run_soap("innerFooGetFancy")
    print output
    assert """<innerFooGetFancyResponse><result xsi:type="FancyValue"><age xsi:type="xsd:int">42</age><computed xsi:type="xsd:string">Hello!</computed><name xsi:type="xsd:string">Mr. Bean</name></result></innerFooGetFancyResponse>""" in output

def test_documentation_strings():
    cherrypy.root = MyService("http://foo.bar.baz")
    testutil.create_request("/soap/api.wsdl")
    output = cherrypy.response.body[0]
    print "WSDL:\n\n"
    print output
    root = et.fromstring(output)
    porttype = root.findall(".//{http://schemas.xmlsoap.org/wsdl/}portType")[0]
    operation = porttype.getchildren()[0]
    assert operation.attrib["name"] == "times2"
    docelems = operation.findall(
                ".//{http://schemas.xmlsoap.org/wsdl/}documentation")
    assert len(docelems) == 1
    assert docelems[0].text == "Multiplies value by two."

def test_fault_for_invalid_input():
    cherrypy.root = MyService("http://foo.bar.baz")
    output = run_soap("times2", """<value>whaleblubber</value>""")
    print cherrypy.response.status
    print output
    assert cherrypy.response.status == "500 Invalid Input"
    assert output == """<Envelope xmlns="http://schemas.xmlsoap.org/soap/envelope/"><Body><Fault><faultcode>Client</faultcode><faultstring>whaleblubber value for the 'value' parameter is not a valid int</faultstring></Fault></Body></Envelope>"""

def test_fault_for_exception():
    cherrypy.root = MyService("http://foo.bar.baz")
    output = run_soap("twentyover", """<value>0</value>""")
    print cherrypy.response.status
    assert cherrypy.response.status == "500 Internal Server Error"
    print output
    assert "SOAP method: twentyover" in output
    assert "foo.bar.baz/soap/" in output, "Raw request should be in the fault"
    assert "return 20 / value" in output, "Traceback should be in fault"
    assert "ZeroDivisionError" in output, "error should be in fault"

def test_boolean_type():
    elem = soap.soap_value("test", bool, True)
    output = str(elem)
    print output
    assert output == '<test xsi:type="xsd:boolean">true</test>'

def test_big_number():
    elem = soap.soap_value("test", int, 3063485326L)
    output = str(elem)
    assert output == '<test xsi:type="xsd:int">3063485326</test>'

def test_empty_list():
    cherrypy.root = ComplexService("http://foo.bar.baz")
    output = run_soap("getempty")
    print output
    assert "fault" not in output
    assert '<result xsi:type="FancyValue_Array"/>' in output

def test_too_many_arguments():
    output = run_soap("somestrings", "<bogus>value</bogus>")
    print output
    root = et.fromstring(output)
    fc = root.findall(".//{http://schemas.xmlsoap.org/soap/envelope/}faultcode")
    assert len(fc) == 1
    fc = fc[0]
    print "Fault code: ", fc.text
    assert fc.text == "Client"

def test_bogus_function_gives_client_fault():
    output = run_soap("nonexistentfunction")
    print output
    root = et.fromstring(output)
    fc = root.findall(".//{http://schemas.xmlsoap.org/soap/envelope/}faultcode")
    assert len(fc) == 1
    fc = fc[0]
    print "Fault code: ", fc.text
    assert fc.text == "Client"

def test_subclass_complex_types_should_appear_in_wsdl():
    output = get_wsdl()
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}complexType")
    foundperson = False
    for elem in schema_elements:
        if "name" in elem.attrib and elem.attrib["name"] == "SubFood":
            for child in elem.getchildren()[0].getchildren():
                if child.attrib["name"] == "person":
                    foundperson = True
                    print "SubFood person type is %s" % child.attrib["type"]
                    assert child.attrib["type"] == "types:FancyValue", \
                              "SubFood person should be a FancyValue"
    assert foundperson, "Should have found the person attribute on SubFood"

def test_unsigned_integer():
    output = get_wsdl()
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}complexType")
    founditem = False
    for elem in schema_elements:
        if "name" in elem.attrib and elem.attrib["name"] == "FoodItem":
            for child in elem.getchildren()[0].getchildren():
                if child.attrib["name"] == "quantity":
                    founditem = True
                    print "FoodItem quantity type is %s" % child.attrib["type"]
                    assert child.attrib["type"] == "xsd:unsignedInt"
    assert founditem, "Should have found FoodItem's quantity field"
    elem = soap.soap_value("quantity", runtime.unsigned, 5)
    assert str(elem) == '<quantity xsi:type="xsd:unsignedInt">5</quantity>'

class NamespacedRoot(WebServicesRoot):
    @wsexpose(str)
    def world(self):
        return "Hello!"

def test_explicit_namespace():
    cherrypy.root = NamespacedRoot("https://foo.bar.baz/", 
        "http://foo.bar/service/URI/")
    output = get_wsdl()
    print output
    assert 'targetNamespace="http://foo.bar/service/URI/"' in output
    assert 'xmlns:types="http://foo.bar/service/URI/types"' in output

class OptionalParams(WebServicesRoot):
    @wsexpose(int)
    @wsvalidate(int, int, int)
    def addem(self, num1, num2=2, num3=3):
        return num1+num2+num3

def test_optional():
    cherrypy.root = OptionalParams("https://foo.bar.baz")
    output = get_wsdl()
    print output
    root = et.fromstring(output)
    schema_elements = root.findall(
                            ".//{http://www.w3.org/2001/XMLSchema}element")
    for elem in schema_elements:
        if elem.attrib.get("name", None) == "addem":
            complextype = elem.getchildren()[0]
            sequence = complextype.getchildren()[0]
            num1 = sequence.getchildren()[0]
            assert num1.attrib["name"] == "num1"
            assert "minOccurs" not in num1.attrib
            num2 = sequence.getchildren()[1]
            assert num2.attrib["name"] == "num2"
            assert num2.attrib["minOccurs"] == "0"
            num3 = sequence.getchildren()[2]
            assert num3.attrib["name"] == "num3"
            assert num3.attrib["minOccurs"] == "0"

class StringRoot(WebServicesRoot):
    @wsexpose(str)
    @wsvalidate(str)
    def say_hello(self, name):
        assert name is not None
        assert name != "None"
        return "Hi, %s" % name

def test_empty_parameter_is_empty_string():
    cherrypy.root = StringRoot("http://foo.bar.baz/")
    output = run_soap("say_hello", """<name></name>""")
    print output
    assert "fault" not in output

class Person(object):
    name = str
    age = int

class ComplexInput(WebServicesRoot):
    @wsexpose()
    @wsvalidate(Person)
    def savePerson(self, p):
        self.person = p
    
def test_complex_input_simple_object():
    ci = ComplexInput("http://foo.bar.baz/")
    cherrypy.root = ci
    output = run_soap("savePerson", """<p>
    <name>George</name>
    <age>280</age>
</p>""")
    print output
    assert isinstance(ci.person, Person)


"""A template plugin that generates simple XML"""
from genshi.builder import Element

import dispatch

from tgwebservices.runtime import primitives, ctvalues, typedproperty

_simplevalues = set([int, long, basestring, str, unicode, float, bool])

@dispatch.generic()
def xml_value(name, value):
    """Converts a value into XML Element objects."""
    pass

@xml_value.when("type(value) in _simplevalues")
def xml_simple(name, value):
    elem = Element(name)
    elem(str(value))
    return elem

@xml_value.when("type(value) not in primitives")
def xml_instance(name, value):
    """Handles an instance of a complex type."""
    elem = Element(name)
    cls = value.__class__
    for key in ctvalues(cls):
        elem.append(xml_value(key, getattr(value, key)))
    return elem

@xml_value.when("isinstance(value, list)")
def xml_list(name, value):
    elem = Element(name)
    for item in value:
        elem.append(xml_value("item", item))
    return elem

@xml_value.when("isinstance(value, dict)")
def xml_dict(name, value):
    elem = Element(name)
    for key in value.keys():
        if key.startswith("tg_"):
            continue
        elem.append(xml_value(key, value[key]))
    return elem

class AutoXMLTemplate(object):

    def __init__(self, extra_vars_func=None, options=None):
        pass

    def load_template(self, templatename):
        "There are no actual templates with this engine"
        pass

    def render(self, info, format="html", fragment=False, template=None):
        "Renders the template to a string using the provided info."
        if "result" not in info:
            info = dict(result=info)
        return str(xml_value("result", info["result"]))

    def get_content_type(self, user_agent):
        return "text/xml"


"""Input Conversion Routines"""

import re

import simplejson
from turbogears import validators

from tgwebservices.runtime import primitives

namespace_expr = re.compile(r'^\{.*\}')

def boolean_converter(value):
    value = value.lower()
    if value in ["", "false", "0", "no", "off", "f", "n"]:
        return False
    if value in ["true", "1", "yes", "on", "y", "t"]:
        return True
    raise validators.Invalid("%s is not a legal boolean value" % (value),
                             value, None)

def _get_single_value(elem, itype):
    """Converts a single XML element into the given type"""
    if isinstance(itype, list):
        itemtype = itype[0]
        items = []
        for subelem in elem.getchildren():
            items.append(_get_single_value(subelem, itemtype))
        return items
    if not isinstance(itype, type):
        itype = type(itype)
    if itype not in primitives:
        return _xml_to_instance(elem, itype)
    if itype == bool:
        itype = boolean_converter
        
    if elem.text is None:
        text = ""
    else:
        text = elem.text
        
    try:
        return itype(text)
    except ValueError:
        raise validators.Invalid(
            "%s value for the '%s' parameter is not a "
            "valid %s" % (text, namespace_expr.sub("", elem.tag),
                         itype.__name__), 
            text, None)
    
def _xml_to_instance(input, cls):
    """Converts an input element into a new instance of cls."""
    instance = cls()
    for elem in input.getchildren():
        tag = namespace_expr.sub("", elem.tag)
        try:
            itype = getattr(cls, tag)
        except AttributeError:
            raise validators.Invalid("%s is an unknown tag for a %s" 
                                     % (tag, cls.__name__),
                tag, None)
        setattr(instance, tag, _get_single_value(elem, itype))
    return instance

def handle_xml_params(body, input_types):
    kw = {}
    for elem in body.getchildren():
        param = namespace_expr.sub("", elem.tag)
        try:
            itype = input_types[param]
        except KeyError:
            raise validators.Invalid(
                "%s is not a valid parameter (valid values are: %s)"
                % (param, input_types.keys()), param, None
            )
        kw[param] = _get_single_value(elem, itype)
    return kw

def handle_keyword_params(kw, input_types):
    # convert the input parameters to appropriate types
    for key in kw:
        if key in input_types:
            try:
                converter = input_types[key]
                if converter == bool:
                    converter = boolean_converter
                kw[key] = converter(kw[key])
            except ValueError:
                raise validators.Invalid(
                    "%s value for the '%s' parameter is not a "
                    "valid %s" % (kw[key], key, 
                                 input_types[key].__name__), 
                    kw[key], None)
        else:
            raise validators.Invalid(
                "%s is not a valid parameter (valid values are: %s)"
                % (key, input_types.keys()), key, None
            )

def _get_json_value(value, itype):
    if isinstance(itype, list):
        itemtype = itype[0]
        return [_get_json_value(item, itemtype) for item in value]
    elif not isinstance(itype, type):
        return _get_json_value(value, type(itype))
    elif itype not in primitives and isinstance(value, dict):
        return _create_instance_from_json(value, itype)
    return itype(value)

def _create_instance_from_json(value, cls):
    instance = cls()
    for key in value:
        try:
            itype = getattr(cls, key)
        except AttributeError:
            raise validators.Invalid("%s is an unknown parameter for a %s" 
                                     % (key, cls.__name__),
                key, None)
        setattr(instance, str(key), _get_json_value(value[key], itype))
    return instance

def handle_json_params(input, input_types):
    if isinstance(input, basestring):
        try:
            input = simplejson.loads(input)
        except ValueError:
            raise validators.Invalid("Invalid JSON input")
    kw = {}
    for key in input:
        if key in input_types:
            kw[str(key)] = _get_json_value(input[key], input_types[key])
        else:
            raise validators.Invalid(
                "%s is not a valid parameter (valid values are: %s)"
                % (key, input_types.keys()), key, None
            )
    return kw
        



from tgwebservices.controllers import WebServicesRoot, WebServicesController, \
                                      wsexpose, wsvalidate

from tgwebservices.runtime import typedproperty, unsigned

class MyService(WebServicesRoot):
    @wsexpose(int)
    @wsvalidate(int)
    def times2(self, value):
        "Multiplies value by two."
        return value * 2
    
    @wsexpose(int)
    @wsvalidate(int)
    def twentyover(self, value):
        "Divides 20 by value"
        return 20 / value
    
class FancyValue(object):
    name = ""
    age = int
    
    def computed(self):
        return "Hello!"
    computed = typedproperty(str, computed)
    
    def __init__(self, name=None, age=None):
        self.name = name
        self.age = age

class ComplexProperty(object):
    def athing(self):
        return FancyValue("Arnie", 55)
    athing = typedproperty(FancyValue, athing)

class FoodItem(object):
    person = FancyValue
    quantity = unsigned
    
    def __init__(self, name=None, quantity=1):
        self.person = FancyValue(name)
        self.quantity = quantity

class SubFood(FoodItem):
    price = 1.95

class FoodOrder(object):
    name = ""
    items = [FoodItem]

class ComplexService(WebServicesRoot):
    
    @wsexpose(FancyValue)
    def getfancy(self):
        "Returns a fancy value"
        fv = FancyValue()
        fv.name = "Mr. Test"
        fv.age = 33
        return fv
    
    @wsexpose(SubFood)
    def getsub(self):
        sf = SubFood()
        return sf
    
    @wsexpose([FancyValue])
    def getmulti(self):
        return [FancyValue("Mr. Washington", 274), 
                FancyValue("Mr. Lincoln", 197)]
    
    @wsexpose([FancyValue])
    def getempty(self):
        return []
    
    @wsexpose([str])
    def somestrings(self):
        return ["A", "B", "C"]
    
    @wsexpose(FoodOrder)
    def getorder(self):
        order = FoodOrder()
        order.name = "WHC"
        order.items = [FoodItem("Burger"), FoodItem("Fries", 2)]
        return order
    
    @wsexpose(ComplexProperty)
    def getcomprop(self):
        return ComplexProperty()
    
    @wsexpose(FancyValue)
    @wsvalidate(FancyValue)
    def tenyearsolder(self, person):
        person.age = person.age + 10
        self.last_person = person
        return person

class InnerService(WebServicesController):
    def __init__(self, add_another=False):
        if add_another:
            self.foo = InnerService()
        
    @wsexpose(FancyValue)
    def getFancy(self):
        return FancyValue("Mr. Bean", 42)

class OuterService(WebServicesRoot):
    inner = InnerService(True)
    another = InnerService()

    @wsexpose(int)
    def ultimatequestion(self):
        return 42



"""Implements the SOAP protocol. Specifically, this implementation talks in
the wrapped document/literal form."""

import re
import sys
import traceback
import time
import logging

import dispatch
import cherrypy
import cElementTree as et
from genshi.builder import tag, Element, Namespace

from turbogears import controllers, expose as tgexpose, error_handler, \
                       exception_handler, validators

from tgwebservices.runtime import primitives, ctvalues, typedproperty, unsigned

log = logging.getLogger("tgwebservices.soap")

namespace_expr = re.compile(r'^\{.*\}')

_simplevalues = set([int, long, basestring, str, unicode, float])

@dispatch.generic()
def soap_value(name, return_type, value):
    """Converts a value into xml Element objects for inclusion in the
    SOAP response output."""
    pass

@soap_value.when("type(value) in _simplevalues")
def soapint(name, return_type, value):
    elem = Element(name, **{"xsi:type" : soap_type(return_type)})
    elem(str(value))
    return elem

@soap_value.when("isinstance(value, bool)")
def soapbool(name, return_type, value):
    elem = Element(name, **{"xsi:type" : "xsd:boolean"})
    elem(str(value).lower())
    return elem

@soap_value.when("type(value) not in primitives")
def soap_instance(name, return_type, value):
    """Handles an instance of a complex type."""
    elem = Element(name, **{"xsi:type" : value.__class__.__name__})
    cls = value.__class__
    for key in ctvalues(cls):
        elem.append(soap_value(key, getattr(cls, key),
                    getattr(value, key)))
    return elem

@soap_value.when("isinstance(value, list)")
def soap_list(name, return_type, value):
    # in the context of the SOAP output, the types: namespace is implied
    elem = Element(name, **{"xsi:type" : 
            soap_type(return_type).replace("types:", "")})
    itemtype = soap_type(return_type)
    for item in value:
        elem.append(soap_value("item", itemtype, item))
    return elem

def soap_body(methodname, function_info, output):
    """Generates xml Elements for the body of a SOAP response."""
    body = Element(methodname + "Response")
    body.append(soap_value("result", function_info.return_type, output))
    return body


# Maps Python types to XML Schema types
type_registry = {
    basestring : "xsd:string",
    str : "xsd:string",
    unicode : "xsd:string",
    int : "xsd:int",
    long : "xsd:long",
    float : "xsd:float",
    bool : "xsd:boolean",
    unsigned : "xsd:unsignedInt"
}

# Maps python types to appropriately named array types.
array_registry = {
    basestring : "String_Array",
    str : "String_Array",
    unicode : "String_Array",
    int : "Int_Array",
    long : "Long_Array",
    float : "Float_Array",
    bool : "Boolean_Array",
}

def handle_list(kind, arrays):
    """Figures out what kind of array the user has declared and registers
    that type. Arrays need to be declared separately in the XML schema."""
    # check for primitive types
    if kind in array_registry:
        arraytype = array_registry[kind]
    elif not isinstance(kind, type):
        return handle_list(kind.__class__, arrays)
    else:
        arraytype = kind.__name__ + "_Array"
    if arrays is not None:
        arrays.add(kind)
    return arraytype
        

def soap_type(kind, arrays=None):
    """Figures out an appropriate schema declaration for the object that is 
    passed in."""
    if kind is None:
        return "xsd:string"
    
    if isinstance(kind, typedproperty):
        return soap_type(kind.type, arrays)
        
    if isinstance(kind, list):
        try:
            return "types:" + handle_list(kind[0], arrays)
        except IndexError:
            raise IndexError("Cannot determine SOAP data type from empty list")
    if kind in type_registry:
        return type_registry.get(kind)
    if not isinstance(kind, type):
        return soap_type(kind.__class__, arrays)
    return "types:" + kind.__name__

class SoapController(controllers.Controller):
    """Controller that manages the SOAP requests and responses. This also
    provides the api.wsdl URL that describes the web service."""
    def __init__(self, wscontroller, tns, typenamespace):
        self.wscontroller = wscontroller
        self.arrays = None
        self.tns = tns
        self.typenamespace = typenamespace
    
    def _cp_on_error(self):
        e = sys.exc_info()[1]
        soap = Namespace("http://schemas.xmlsoap.org/soap/envelope/")
        
        # set up elements common to all faults
        envelope = Element(soap.Envelope)
        body = Element("Body")
        envelope.append(body)
        fault = Element("Fault")
        body.append(fault)
        faultcode = Element("faultcode")
        fault.append(faultcode)
        faultstring = Element("faultstring")
        faultstring.append(str(e))
        fault.append(faultstring)
        
        cherrypy.response.headers["Content-Type"] = "text/xml;charset=utf-8"
        
        # customize the fault for the type of problem
        if isinstance(e, validators.Invalid):
            cherrypy.response.status = "500 Invalid Input"
            faultcode.append("Client")
        else:
            cherrypy.response.status = "500 Internal Server Error"
            faultcode.append("Server")
            detail = Element("detail")
            fault.append(detail)
            method = getattr(cherrypy.request, "soap_method", "")
            data = getattr(cherrypy.request, "soap_start", "")
            detail.append("Error occurred at %s\n" % (time.asctime()))
            detail.append("SOAP method: %s\n" % method)
            detail.append("\nRaw request (first 2K):\n%s\n" % data)
            detail.append("\nTraceback:\n%s\n" % 
                "\n".join(traceback.format_exception(*sys.exc_info())))
            
        cherrypy.response.body = str(envelope)
    
    @tgexpose("genshi:tgwebservices.templates.soap",
              content_type="text/xml;charset=utf-8", format="xml")
    def index(self):
        """Processes SOAP requests and generates the responses."""
        # get request data and produce an ElementTree that we can work with.
        request = cherrypy.request
        response = cherrypy.response
        clen = int(request.headers.get('Content-Length')) or 0
        data = request.body.read(clen)
        request.soap_start = data[:2048]
        soapreq = et.fromstring(data)
        
        # find the body of the request and the specific method name that has
        # been requested.
        body = soapreq.find("{http://schemas.xmlsoap.org/soap/envelope/}Body")
        body = body.getchildren()[0]
        methodname = body.tag
        methodname = namespace_expr.sub("", methodname)
        request.soap_method = methodname
        
        try:
            method = self.wscontroller._ws_funcs[methodname]
        except KeyError:
            raise validators.Invalid("Unknown SOAP method: %s" %
                                     methodname, methodname, None)
        
        params = {"_tgws" : True}
        params["xml_body"] = body
        
        # ensure that the requested method has actually been exposed
        assert hasattr(method, "exposed") and method.exposed == True
        
        # set the _tgws flag so that the TurboGears wrapping is not applied
        # in wsexpose. We want to manage the produced dictionary here, because
        # we need to pass additional information on to the soap template.
        output = method(**params)
        data = dict(baseURL=self.wscontroller._ws_baseURL,
                    tns=self.tns,
                    typenamespace=self.typenamespace,
                    soap_body=soap_body, output=output,
                    methodname=methodname, 
                    function_info =
                        self.wscontroller._ws_funcs[methodname]._ws_func_info)
        return data
    
    @tgexpose("genshi:tgwebservices.templates.wsdl", format="xml",
              content_type="text/xml;charset=utf-8")
    def api_wsdl(self, xmlsoap=False):
        """Produces the WSDL for this SOAP service."""
        if self.arrays is None:
            self._initialize_arrays()
        ct = list(self.wscontroller._ws_complex_types)
        ct.sort(key=lambda item: item._type_dependents, reverse=True)
        funclist = self.wscontroller._ws_funcs.keys()
        funclist.sort()
        
        # Apache Axis2 demands the w3.org namespace for soapenc
        # Axis1 needs the xmlsoap variety.
        if xmlsoap:
            soapenc = "http://schemas.xmlsoap.org/soap/encoding/"
        else:
            soapenc = "http://www.w3.org/2001/09/soap-encoding"
        return dict(service_name=self.wscontroller.__class__.__name__,
                    baseURL=self.wscontroller._ws_baseURL,
                    registry=self.wscontroller._ws_funcs, soap_type=soap_type,
                    funclist=funclist,
                    soap_array=handle_list,
                    complex_types=ct,
                    ctvalues=ctvalues,
                    arrays=self.arrays, soapenc=soapenc,
                    tns=self.tns, typenamespace=self.typenamespace)
    
    @tgexpose()
    def xmlsoap_wsdl(self):
        return self.api_wsdl(xmlsoap=True)
    
    def _initialize_arrays(self):
        arrays = set()
        for func in self.wscontroller._ws_funcs.values():
            soap_type(func._ws_func_info.return_type, arrays)
        for cls in self.wscontroller._ws_complex_types:
            for key in ctvalues(cls):
                soap_type(getattr(cls, key), arrays)
        self.arrays = arrays



from setuptools import setup, find_packages
from turbogears.finddata import find_package_data

import os
execfile(os.path.join("tgwebservices", "release.py"))

setup(
    name="TGWebServices",
    version=version,
    
    description=description,
    long_description=long_description,
    author=author,
    author_email=email,
    url=url,
    license=license,
    
    install_requires = [
        "TurboGears >= 1.0",
        "Genshi >= 0.3.4"
    ],
    scripts = [],
    zip_safe=False,
    packages=find_packages(),
    package_data = find_package_data(where='tgwebservices',
                                     package='tgwebservices'),
    keywords = [
        "turbogears.extension"
    ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: TurboGears',
        # if this is an application that you'll distribute through
        # the Cheeseshop, uncomment the next line
        # 'Framework :: TurboGears :: Applications',
        
        # if this is a package that includes widgets that you'll distribute
        # through the Cheeseshop, uncomment the next line
        # 'Framework :: TurboGears :: Widgets',
    ],
    test_suite = 'nose.collector',
    entry_points="""
    [python.templating.engines]
    wsautoxml = tgwebservices.xml:AutoXMLTemplate
    """
    )
    



import cElementTree as et

from tgwebservices.iconv import _get_single_value, handle_json_params

class Person(object):
    name = str
    age = int
    married = False

def test_complex_conversion_simple_class():
    tree = et.fromstring("""<p>
    <name>Super Value Menu</name>
    <age>17</age>
    <married>no</married>
</p>""")
    val = _get_single_value(tree, Person)
    print val
    assert isinstance(val, Person)
    assert val.age == 17
    assert val.name == "Super Value Menu"
    print val.married
    assert val.married == False

class Family1(object):
    lastname = str
    members = [str]

def test_complex_conversion_with_list():
    tree = et.fromstring("""<family>
    <lastname>Brady</lastname>
    <members>
        <item>Mike</item>
        <item>Carol</item>
        <item>Greg</item>
        <item>Peter</item>
        <item>Bobby</item>
        <item>Marsha</item>
        <item>Jan</item>
        <item>Cindy</item>
    </members>
</family>
""")
    val = _get_single_value(tree, Family1)
    assert val.lastname == "Brady"
    print val.members
    assert len(val.members) == 8
    assert val.members[2] == "Greg"

class Family2(object):
    lastname = str
    members = [Person]

def test_complex_with_list_of_instances():
    tree = et.fromstring("""<family>
    <lastname>Brady</lastname>
    <members>
        <item>
            <name>Mike</name>
            <age>42</age>
        </item>
        <item>
            <name>Carol</name>
            <age>40</age>
        </item>
        <item>
            <name>Greg</name>
            <age>17</age>
        </item>
        <item>
            <name>Peter</name>
            <age>13</age>
        </item>
        <item>
            <name>Bobby</name>
            <age>8</age>
        </item>
        <item>
            <name>Marsha</name>
            <age>16</age>
        </item>
        <item>
            <name>Jan</name>
            <age>13</age>
        </item>
        <item>
            <name>Cindy</name>
            <age>7</age>
        </item>
    </members>
</family>
""")
    val = _get_single_value(tree, Family2)
    assert val.lastname == "Brady"
    print val.members
    assert len(val.members) == 8

def test_list_of_bools():
    tree = et.fromstring("""<list>
    <item>no</item>
    <item>false</item>
    <item>yes</item>
    <item>False</item>
    <item>true</item>
</list>""")
    val = _get_single_value(tree, [bool])
    print val
    assert len(val) == 5

def test_json_conversion_of_class():
    input_types = {"person" : Person}
    data = {"person" : {"name" : "Fred", "age" : 99}}
    kw = handle_json_params(data, input_types)
    assert isinstance(kw["person"], Person)
    print kw["person"].age
    assert kw["person"].age == 99
    assert kw["person"].name == "Fred"

def test_json_conversion_of_nested_class():
    input_types = {"family" : Family2}
    data = {"family" : {"lastname" : "Brady", "members" :
        [{"name" : "Mike", "age" : 42}, {"name" : "Carol", "age" : 40}]
}}
    kw = handle_json_params(data, input_types)
    assert isinstance(kw["family"], Family2)
    assert len(kw["family"].members) == 2
    assert kw["family"].members[0].name == "Mike"
    assert kw["family"].members[0].age == 42